#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <ssdef.h>
#include <descrip.h>

#include "hiredis.h"

static char *in_cstr(struct dsc$descriptor *s)
{
    int slen, ix;
    char *sptr, *res;
    if(s->dsc$b_class == DSC$K_CLASS_S || s->dsc$b_class == DSC$K_CLASS_D)
    {
        slen = s->dsc$w_length;
        sptr = s->dsc$a_pointer;
    }
    else if(s->dsc$b_class == DSC$K_CLASS_VS) 
    {
        slen = *((short int *)s->dsc$a_pointer);
        sptr = s->dsc$a_pointer + sizeof(short int);
    }
    else
    {
        fprintf(stderr, "Unsupported descriptor class: %d\n", s->dsc$b_class);
        exit(44);
    }
    res = malloc(slen + 1);
    strncpy(res, sptr, slen);
    res[slen] = 0;
    /* trim traling whitespace */
    ix = slen - 1;
    while(ix >= 0 && (res[ix] == ' ' || res[ix] == 0))
    {
        res[ix] = 0;
        ix--;
    }
    return res;
}

static void out_cstr(char *s1, struct dsc$descriptor *s2, short int *len)
{
    char *sptr;
    short int slen;
    slen = strlen(s1);
    if(slen > s2->dsc$w_length) slen = s2->dsc$w_length;
    if(s2->dsc$b_class == DSC$K_CLASS_S || s2->dsc$b_class == DSC$K_CLASS_D)
    {
        memcpy(s2->dsc$a_pointer, s1, slen);
    }
    else if(s2->dsc$b_class == DSC$K_CLASS_VS)
    {
        memcpy(s2->dsc$a_pointer + sizeof(short int), s1, slen);
        *((short int *)s2->dsc$a_pointer) = slen;
    }
    else
    {
        fprintf(stderr, "Unsupported descriptor class: %d\n", s2->dsc$b_class);
        exit(44);
    }
    *len = slen;
}

int VMS_HIREDIS_CONNECT(struct dsc$descriptor *host, int *port, int *cache_id, long long int *cache_handle)
{
   redisReply *reply;
   char *host2;
   host2 = in_cstr(host);
   *cache_handle = (long long int)redisConnect(host2, *port);
   free(host2);
   if(*cache_handle == 0) return SS$_ABORT;
   reply = redisCommand((redisContext *)(*cache_handle), "SELECT %d", *cache_id);
   if(reply == NULL) return SS$_ABORT;
   freeReplyObject(reply);   
   return SS$_NORMAL;
}

int VMS_HIREDIS_SET_STRING(long long int *cache_handle, struct dsc$descriptor *key, struct dsc$descriptor *val)
{
   redisReply *reply;
   char *key2, *val2;
   key2 = in_cstr(key);
   val2 = in_cstr(val);
   reply = redisCommand((redisContext *)(*cache_handle), "SET %s %s", key2, val2);
   free(key2);
   free(val2);
   if(reply == NULL) return SS$_ABORT;
   freeReplyObject(reply);   
   return SS$_NORMAL;
}

int VMS_HIREDIS_SET_INTEGER(long long int *cache_handle, struct dsc$descriptor *key, int *val)
{
   redisReply *reply;
   char *key2;
   key2 = in_cstr(key);
   reply = redisCommand((redisContext *)(*cache_handle), "SET %s %b", key2, val, sizeof(int));
   free(key2);
   if(reply == NULL) return SS$_ABORT;
   freeReplyObject(reply);   
   return SS$_NORMAL;
}

int VMS_HIREDIS_GET_STRING(long long int *cache_handle, struct dsc$descriptor *key, struct dsc$descriptor *val, short int *vallen)
{
   redisReply *reply;
   char *key2;
   key2 = in_cstr(key);
   reply = redisCommand((redisContext *)(*cache_handle), "GET %s", key2);
   free(key2);
   if(reply == NULL) return SS$_ABORT;
   out_cstr(reply->str, val, vallen);
   freeReplyObject(reply);   
   return SS$_NORMAL;
}

int VMS_HIREDIS_GET_INTEGER(long long int *cache_handle, struct dsc$descriptor *key, int *val)
{
   redisReply *reply;
   char *key2;
   key2 = in_cstr(key);
   reply = redisCommand((redisContext *)(*cache_handle), "GET %s", key2);
   free(key2);
   if(reply == NULL) return SS$_ABORT;
   *val = *((int *)reply->str);
   freeReplyObject(reply);   
   return SS$_NORMAL;
}

int VMS_HIREDIS_KEYS_LOOKUP(long long int *cache_handle, struct dsc$descriptor *key, long long int *it_handle)
{
   char *key2;
   key2 = in_cstr(key);
   *it_handle = (long long int)redisCommand((redisContext *)(*cache_handle), "KEYS %s", key2);
   free(key2);
   if(*it_handle == 0) return SS$_ABORT;
   return SS$_NORMAL;
}

int VMS_HIREDIS_KEYS_COUNT(long long int *it_handle)
{
   return ((redisReply *)(*it_handle))->elements;
}

void VMS_HIREDIS_KEYS_GET(long long int *it_handle, int *ix, struct dsc$descriptor *key, short int *keylen)
{
   char *p;
   p = ((redisReply *)(*it_handle))->element[*ix - 1]->str;
   out_cstr(p, key, keylen);
}

void VMS_HIREDIS_KEYS_FREE(long long int *it_handle)
{
   freeReplyObject((redisReply *)(*it_handle));
}

void VMS_HIREDIS_DISCONNECT(long long int *cache_handle)
{
   redisFree((redisContext *)(*cache_handle));
}
